// Assembler.cpp : implementation of assembler base class
//
// Copyright (C) 2000 Jeremy Goetsch (jeremy@acz.org)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//

#include "stdafx.h"
#include "Asmide.h"
#include "Assembler.h"
#include "AsmOutputStream.h"
#include "IntermediateFile.h"
#include <iostream.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// Error number definitions

// Miscellaneous/system errors (0-99)
#define ERR_TEST			0
#define ERR_UNKNOWN			1
#define ERR_MESSAGE			2
#define WRN_MESSAGE			3
#define ERR_FILE_EMPTY		4
#define ERR_PLUGIN			5

// Preprocessor errors (100-199)
#define ERR_FILE_OPEN		100
#define ERR_FILE_NOEXIST	101
#define WRN_FILE_INCLUDED	102
#define ERR_LINE_TOOLONG	103
#define ERR_UNKNOWN_PRE_DIR	110
#define ERR_MISSING_FNAME	111
#define ERR_MISSING_MACRO	112
#define ERR_TOOMANY_ARGS	113
#define ERR_INVALID_ARGS	114
#define ERR_ARG_COUNT		115
#define ERR_NO_ENDIF		120
#define ERR_ELSE_NOIF		123
#define ERR_ENDIF_NOIF		124

// Parsing errors (200-299)
#define ERR_MISSING_EXP		200
#define ERR_UNKNOWN_LABEL	201
#define ERR_ILLEGAL_EXPR	202
#define ERR_NO_END_QUOTE	203
#define ERR_MISSING_VALUE	204
#define ERR_MISSING_STRING	205
#define ERR_DIVIDEBYZERO	206
#define ERR_NO_OPENPARENTH	207
#define ERR_NO_CLOSEPARENTH	208
#define ERR_EXPR_TOOLONG	210

// Label/directive errors (300-399)
#define ERR_UNKNOWN_DIR		300
#define WRN_DUP_LABEL		302
#define WRN_LABEL_INST		303
#define ERR_INVALID_POS		310
#define ERR_ADDINSTR		311
#define ERR_PAD_NEGATIVE	312
#define ERR_PLUGIN_DIR		313
#define ERR_MODULE_DIR		314
#define ERR_LOCALCHAR_DIR	315

// Instruction table errors (400-499)
#define ERR_UNKNOWN_INST	400
#define ERR_UNKNOWN_ARG		401
#define ERR_MISSING_DATA	402
#define ERR_REL_JUMP		403
#define ERR_INVALID_BIT		404
#define ERR_OVERFLOW		405
#define WRN_MSBYTE_UNUSED	406
#define WRN_ILLEGAL_IND		407

// size of symbol hash table
#define SYMTABLE_HASHSIZE	2237


/////////////////////////////////////////////////////////////////////////////
// CAssembler constructor

CAssembler::CAssembler(UINT nFlags, LPCSTR szDefines)
{
	m_pLabelMap = NULL;
	m_pLabelDisp = NULL;
	m_pInterSrc = NULL;
	m_pOutput = new CAsmOutputStream;
	m_nFlags = nFlags;
	m_pszDefines = szDefines;

	// initialize values
	m_Status = asmOK;
	m_numFiles = 0;
	m_nCodeSize = 0;
	m_nDataSize = 0;
	m_nWarnings = 0;
	m_nErrors = 0;
}

CAssembler::~CAssembler()
{
	if (m_pLabelMap)
		delete m_pLabelMap;
	if (m_pInterSrc)
		delete m_pInterSrc;
}

// SetOutput function
// links the assembler to a CAsmOutputStream object (must be called before Build)
// the object is deleted automatically by the assembler on destruction
/*
void CAssembler::SetOutput(CAsmOutputStream *pOutput)
{
	ASSERT(m_pOutput == NULL);
	ASSERT(pOutput != NULL);
	m_pOutput = pOutput;
}
*/

/////////////////////////////////////////////////////////////////////////////
// Instruction table routines

BOOL CAssembler::LoadTabFile( CString filename)
{
	CStdioFile fl;
	LPCSTR lpszDir = SetDirDefault();
	if (!fl.Open( filename, CFile::modeRead | CFile::typeText))
	{
		cout << "Cannot open table file: " << (LPCSTR)filename << endl;
		exit(1);
	}
	SetCurrentDirectory(lpszDir);

	char lbuff[ MAX_TAB_LINE ];
	do {
		fl.ReadString( lbuff, MAX_TAB_LINE-1);
	} while (!isalpha(lbuff[0]) && lbuff[0] != '"');

	if (lbuff[0] != '"')
	{
		cout << (LPCSTR)filename << " is not a valid TAB file." << endl;
		exit(1);
	}
#ifdef _DEBUG
	lbuff[25] = '\0';
	TRACE1("Opened table: '%s'\n", lbuff+1);
#endif

	CWaitCursor wait;
	while (fl.ReadString( lbuff, MAX_TAB_LINE-1) != NULL) {
		if (isalpha(lbuff[0])) {
			if (!AddTabEntry( lbuff))
				return FALSE;
		}
	}
	TRACE2("Read %d table entries from %s\n", m_Table.GetCount(), filename);
	return TRUE;
}

BOOL CAssembler::AddTabEntry( char *EntryLine)
{
	int nToks;
	char rule[8];
	UINT tclass;
	TableEntry entry;

	if (EntryLine == NULL)
		return( FALSE);
	strupr( EntryLine);
	nToks = sscanf( EntryLine, "%7s %15s %lx %hi %s %hu %hx %hx",
		entry.instr, entry.arg, &entry.opcode, &entry.nbytes, rule, &tclass, &entry.shift, &entry.or);

	if (nToks < 6)
	{
		cout << "Incomplete entry in table file" << endl;
		return FALSE;
	}
	if (entry.nbytes > 6 || entry.nbytes < 1)
	{
		cout << "Invalid value in nbytes field of table entry" << endl;
		return FALSE;
	}

	if (nToks < 7) entry.shift = 0;
	if (nToks < 8) entry.or = 0;
	if (!strcmp(entry.arg, "\"\""))
		entry.arg[0] = '\0';

	entry.opbytes = entry.opcode > 0x00FF ? 2 : 1;
	entry.argbytes = entry.nbytes - entry.opbytes;
	if (entry.argbytes > 2) {
		entry.opbytes = entry.nbytes - 2;
		entry.argbytes = 2;
	}

	if (!strcmp(rule, "ZBIT"))
		entry.modop = TableEntry::OP_ZBIT;
	else if (!strcmp(rule, "R1"))
		entry.modop = TableEntry::OP_R1;
	else if (!strcmp(rule, "COMBINE") || !strcmp(rule, "ZIX"))
		entry.modop = TableEntry::OP_COMBINE;
	else if (!strcmp(rule, "CSWAP"))
		entry.modop = TableEntry::OP_CSWAP;
	else
		entry.modop = TableEntry::OP_NOTOUCH;
	entry.tclass = (BYTE)tclass;

	if (entry.tclass == 1)
		m_Table.AddTail( entry);
	return TRUE;
}


void CAssembler::LoadFirstPass( CAssembler *pSourceAsm)
{
	if (m_pLabelMap)
		delete m_pLabelMap;
	if (m_pInterSrc)
		delete m_pInterSrc;
	m_pLabelMap = pSourceAsm->m_pLabelMap;
	pSourceAsm->m_pLabelMap = NULL;
	m_pInterSrc = pSourceAsm->m_pInterSrc;
	pSourceAsm->m_pInterSrc = NULL;
}


/*
void CAssembler::BuildAndSend( CString filename)
{
	CString params = BuildAll(filename);
	if (m_Status == asmOK) {
		params = '"' + params + '"';
		if (GetOptions()->m_calcExplrNoMessage)
			params += " /q";
		if ((int)ShellExecute( AfxGetMainWnd()->m_hWnd, NULL, GetOptions()->m_calcExplrPath,
				params, GetDirFromPath(filename), SW_SHOW) <= 32)
			AfxMessageBox( IDP_ERROR_EXEC, MB_ICONEXCLAMATION);
	}
}

void CAssembler::BuildObj( CString filename)
{
	CAsmProgressDlg statusdlg;
	statusdlg.Create();
	m_pOutput->PrintOutput("Assembling " + filename);

	// preprocess
	CIntermediateFile interSrc;
	Preprocess( filename, interSrc, &statusdlg);
	m_numLines = interSrc.GetNumLines();
	if (statusdlg)
		statusdlg.SetRange( m_numLines * 2);

	// first pass
	if (m_Status <= asmErrors)
		AssembleFirst( interSrc, &statusdlg);

	// second pass
	if (m_Status <= asmWarnings) {
		CString fn;
		if (filename.Find('.') != -1)
			fn = filename.Left(filename.ReverseFind('.')) + ".obj";
		else
			fn = filename + ".obj";
		CFile out;
		if (!out.Open( fn, CFile::modeCreate | CFile::modeWrite))
			AfxMessageBox("Cannot write to file: " + fn, MB_ICONEXCLAMATION);
		else
			AssembleMain( interSrc, out, &statusdlg);
	}

	Finish();
}
*/

/////////////////////////////////////////////////////////////////////////////
// Assembler implementation routines
/////////////////////////////////////////////////////////////////////////////

static char szTokSep[] =  " \t\n";
static char szLineSep[] = "\n";

/////////////////////////////////////////////////////////////////////////////
// CMacroDef routines

CMacroDef::CMacroDef()
{
	Empty();
}

CMacroDef::CMacroDef( LPCSTR szDefinition)
{
	Empty();
	m_szDefinition = szDefinition;
}

void CMacroDef::Empty()
{
	m_NumArgs = 0;
	m_szDefinition.Empty();
	for (int i = 0; i < MAX_MACRO_ARGS; i++)
		m_szArgs[i].Empty();
}

/////////////////////////////////////////////////////////////////////////////
// Preprocess: Run preprocessor and create intermediate file

void CAssembler::Preprocess( CString filename, CMacroMap *pMacroMap)
{
	char linestor[MAX_LINE_LENGTH];	// current line storage
	char lbuff   [MAX_LINE_LENGTH];	// copy of line used for tokenization
	int line = 0;
	char *tok;
	char *arg;
	int nDepth = 0;
	int nChangeLevel;
	int bUseLine = 1;	//  1 = condition true
						//  0 = condition false
	m_strModule = "";
	m_cLocalLabelChar = '@';

	CMacroDef def;
	char *expos;
	char *exptok;
	LPCSTR defptr;
	int nArgs;
	char *args[MAX_MACRO_ARGS];
	LPCSTR argptr;
	int nParenthCount;

	if (!pMacroMap) {
		if (m_pInterSrc != NULL)
			return;
		m_pInterSrc = new CIntermediateFile;
	}

	CStdioFile src;
	if (!src.Open( filename, CFile::modeRead | CFile::typeText)) {
		cout << "Error opening file " << (LPCSTR)filename << endl;
		exit(1);
	}
	m_numFiles++;

	// macro definition table
	BOOL bDestroyMap = FALSE;
	if (!pMacroMap) {
		pMacroMap = new CMacroMap;
		bDestroyMap = TRUE;

		// add predefined macros
		pMacroMap->SetAt("__ASMIDE_VER", CMacroDef( STR_ASMIDE_VER));
		pMacroMap->SetAt("__FILE__", CMacroDef("\"" + src.GetFileName() + "\""));
		pMacroMap->SetAt("__DATE__", CMacroDef( CTime::GetCurrentTime().Format("\"%x\"")));
		pMacroMap->SetAt("__DATELONG__", CMacroDef( CTime::GetCurrentTime().Format("\"%#x\"")));
		pMacroMap->SetAt("__TIME__", CMacroDef( CTime::GetCurrentTime().Format("\"%X\"")));

		if (m_pszDefines != NULL) {
			LPCSTR pDef = m_pszDefines;
			while (!__iscsymf(*pDef) && *pDef != '\0')
				pDef++;
			while (*pDef != '\0') {
				tok = linestor;
				while (__iscsym(*pDef))
					*(tok++) = *(pDef++);
				*tok = '\0';
				pMacroMap->SetAt(linestor, CMacroDef(""));
				while (!__iscsymf(*pDef) && *pDef != '\0')
					pDef++;
			}
		}
	}

	int fMap = m_pInterSrc->AddFileName( src.GetFilePath());

	try {
	 while (src.ReadString( linestor, MAX_LINE_LENGTH-1) != NULL) {
		line++;
		if (m_pOutput->CheckCancel())
			m_Status = asmCanceled;

		// cut off comments
		arg = linestor;
		while (*arg) {
			// semicolon is literal in quotes
			if (*arg == '"') {
				do arg++;
				while (*arg && *arg != '"');
			}
			else if (*arg == '\'')
				arg++;
			else if (*arg == ';') {
				*arg = '\n';
				*(arg+1) = '\0';
				break;
			}
			arg++;
		}

		// check for line too long
		if (strlen(linestor) == MAX_LINE_LENGTH) {
			AddError( ERR_LINE_TOOLONG, src.GetFilePath(), line);
			linestor[strlen(linestor)-1] = '\n';
		}

		strncpy( lbuff, linestor, MAX_LINE_LENGTH-1);
		tok = strtok( lbuff, szTokSep);

		// pre-expansion directives
		if (lbuff[0] == '#') {
			if (!_stricmp(tok, "#IFDEF")) {
				nDepth++;
				if (bUseLine) {
					arg = strtok(NULL, szTokSep);
					if (arg == NULL)
						AddError( ERR_MISSING_MACRO, src.GetFilePath(), line);
					else if (!pMacroMap->Lookup( arg, def)) {
						bUseLine = 0;
						nChangeLevel = nDepth;
					}
				}
				continue;
			}
			else if (!_stricmp(tok, "#IFNDEF")) {
				nDepth++;
				if (bUseLine) {
					arg = strtok(NULL, szTokSep);
					if (arg == NULL)
						AddError( ERR_MISSING_MACRO, src.GetFilePath(), line);
					else if (pMacroMap->Lookup( arg, def)) {
						bUseLine = 0;
						nChangeLevel = nDepth;
					}
				}
				continue;
			}
			else if (bUseLine) {
				if (!_stricmp(tok, "#DEFINE")) {
					tok = strtok( NULL, szLineSep);
					if (tok == NULL) {
						AddError( ERR_MISSING_MACRO, src.GetFilePath(), line);
						continue;
					}
					while (isspace(*tok))
						tok++;
					CString dname;
					while (__iscsym(*tok) && *tok) {
						dname += *tok;
						tok++;
					}
					if (dname.IsEmpty()) {
						AddError( ERR_MISSING_MACRO, src.GetFilePath(), line);
						continue;
					}

					def.Empty();
					if (*tok == '(') {	// macro arguments
						tok++;
						while (isspace(*tok))
							tok++;
						while (*tok != ')' && *tok != '\0') {
							if (def.m_NumArgs >= MAX_MACRO_ARGS) {
								AddError( ERR_TOOMANY_ARGS, src.GetFilePath(), line, dname);
								break;
							}
							while (__iscsym(*tok)) {
								def.m_szArgs[def.m_NumArgs] += *tok;
								tok++;
							}
							def.m_NumArgs++;
							while (isspace(*tok))
								tok++;
							if (*tok == ',') {
								tok++;
								while (isspace(*tok))
									tok++;
							}
							else if (*tok != ')') {
								AddError( ERR_INVALID_ARGS, src.GetFilePath(), line, dname);
								break;
							}
						}
						if (*tok != ')')
							continue;
						tok++;
					}
					while (isspace(*tok))
						tok++;
					def.m_szDefinition = tok;
					pMacroMap->SetAt(dname, def);	// add macro definition
					continue;
				}
				else if (!_stricmp(tok, "#UNDEF")) {
					tok = strtok(NULL, szTokSep);
					if (tok == NULL) {
						AddError( ERR_MISSING_MACRO, src.GetFilePath(), line);
						continue;
					}
					pMacroMap->RemoveKey( tok);	// remove macro definition
					continue;
				}
			}
		}

		// expand macros
		expos = lbuff;
		tok = linestor;
		while (*tok && expos < lbuff+MAX_LINE_LENGTH-1) {
			if (__iscsym(*tok)) {
				exptok = expos;
				while (__iscsym(*tok) && *tok && expos < lbuff+MAX_LINE_LENGTH-1) {
					*expos = *tok;
					expos++;
					tok++;
				}
				*expos = '\0';
				if (pMacroMap->Lookup( exptok, def)) {
					expos = exptok;
					nArgs = 0;
					if (*tok == '(') {	// get args
						nParenthCount = 1;
						for (;;) {
							tok++;
							while (isspace(*tok))
								tok++;
							args[nArgs] = tok;
							nArgs++;
							while (*tok != ',' && *tok) {
								if (*tok == '(')
									nParenthCount++;
								else if (*tok == ')' && !--nParenthCount)
									break;
								tok++;
							}
							if (*tok == ',') {
								*tok = '\0';
							}
							else if (*tok == ')') {
								*tok = '\0';
								tok++;
								break;
							}
							else {
								AddError( ERR_INVALID_ARGS, src.GetFilePath(), line, exptok);
								break;
							}
						}
					}
					if (nArgs != def.m_NumArgs)
						AddError( ERR_ARG_COUNT, src.GetFilePath(), line, exptok);
					else {
						defptr = def.m_szDefinition;
						while (*defptr && expos < lbuff+MAX_LINE_LENGTH-1) {
							if (__iscsym(*defptr)) {
								exptok = expos;
								while (__iscsym(*defptr) && *defptr) {
									*expos = *defptr;
									expos++;
									defptr++;
								}
								*expos = '\0';
								for (int i = 0; i < nArgs; i++) {
									if (!strcmp( exptok, def.m_szArgs[i])) {
										expos = exptok;
										argptr = args[i];
										while (*argptr && expos < lbuff+MAX_LINE_LENGTH-1) {
											*expos = *argptr;
											expos++;
											argptr++;
										}
										break;
									}
								}
							}
							else {
								*expos = *defptr;
								expos++;
								defptr++;
							}
						}
					}
				}
			}
			else if (*tok == '"') {
				// don't expand within quote
				do {
					*expos = *tok;
					expos++;
					tok++;
				} while (*tok != '"' && *tok);
			}
			else {
				*expos = *tok;
				expos++;
				tok++;
			}
		}
		*expos = '\0';
		if (expos == lbuff+MAX_LINE_LENGTH-1)
			AddError( ERR_LINE_TOOLONG, src.GetFilePath(), line);
		strncpy( linestor, lbuff, MAX_LINE_LENGTH-1);

		// get first token
		tok = strtok( lbuff, szTokSep);
		if (tok == NULL)
			continue;
		else if (!_stricmp(tok, ".END"))
			break;

		// post-expansion directives
		if (lbuff[0] == '#') {
			if (!_stricmp(tok, "#IF")) {
				nDepth++;
				if (bUseLine) {
					arg = strtok(NULL, szLineSep);
					if (arg == NULL)
						AddError( ERR_MISSING_EXP, src.GetFilePath(), line);
					else if (!EvalExpr( arg, src.GetFilePath(), line)) {
						bUseLine = 0;
						nChangeLevel = nDepth;
					}
				}
			}
			else if (!_stricmp(tok, "#ELSE")) {
				if (!nDepth)
					AddError( ERR_ELSE_NOIF, src.GetFilePath(), line);
				else if (bUseLine || nDepth == nChangeLevel) {
					bUseLine = !bUseLine;
					if (!bUseLine)
						nChangeLevel = nDepth;
				}
			}
			else if (!_stricmp(tok, "#ENDIF")) {
				if (!nDepth)
					AddError( ERR_ENDIF_NOIF, src.GetFilePath(), line);
				else {
					if (!bUseLine && nDepth == nChangeLevel)
						bUseLine = TRUE;
					nDepth--;
				}
			}

			else if (bUseLine && !_stricmp(tok, "#INCLUDE")) {
				arg = strtok(NULL, szLineSep);
				if (arg == NULL) {
					AddError( ERR_MISSING_FNAME, src.GetFilePath(), line);
					continue;
				}
				CString fn( arg);
				fn.TrimLeft();
				fn.TrimRight();
				// check for quotes
				if (fn.Left(1) == "\"" && fn.Right(1) == "\"")
					fn = fn.Mid( 1, fn.GetLength() - 2);

				CFileStatus fstat;
				// check current directory
				if (!CFile::GetStatus( fn, fstat)) {
					// check include path
					LPCSTR lpszDir = SetDirDefault();
					fn = GetOptions()->m_asmIncludePath + "\\" + fn;
					if (!CFile::GetStatus( fn, fstat)) {
						// not found
						SetCurrentDirectory(lpszDir);
						AddError( ERR_FILE_NOEXIST, src.GetFilePath(), line, arg);
						continue;
					}
					fn = fstat.m_szFullName;
					SetCurrentDirectory(lpszDir);
				}
				// check if already included
				if (m_pInterSrc->IsFileDefined( fstat.m_szFullName)) {
					AddError( WRN_FILE_INCLUDED, filename, line, arg);
					continue;
				}
				Preprocess( fn, pMacroMap);
			}
			else if(bUseLine)
				AddError( ERR_UNKNOWN_PRE_DIR, src.GetFilePath(), line, tok);
			continue;
		}
		if (m_Status > asmErrors)
			break;

		if (bUseLine) {
			// replace backslashes with newlines
			arg = linestor;
			while (*arg) {
				if (*arg == '\\') {
					*arg = '\n';
					m_pInterSrc->AddLine( fMap, line);
				}
				else if (*arg == '"') {
					do
					{
						arg++;
						if(*arg == '\\' && *(arg+1) == '"')
							arg += 2;
					} while (*arg != '"' && *arg);
				}
				else if (*arg == '\'') {
					do
					{
						arg++;
						if(*arg == '\\' && *(arg+1) == '\'')
							arg += 2;
					} while (*arg != '\'' && *arg);
				}
				arg++;
			}

			// write the string
			m_pInterSrc->WriteString( linestor);
			m_pInterSrc->AddLine( fMap, line);
		}
	} }

	catch (CFileException *e) {
		m_Status = asmFileException;
		e->ReportError();
	}
	catch (CMemoryException *e) {
		m_Status = asmMemException;
		e->ReportError();
	}

	if (nDepth)
		AddError( ERR_NO_ENDIF, src.GetFilePath(), line);

	if (bDestroyMap) {
		// make sure there's at least one line of input
		if (m_pInterSrc->GetLength() == 0) {
			m_pInterSrc->WriteString("\n");
			m_pInterSrc->AddLine( fMap, 0);
		}
		delete pMacroMap;
	}
}


/////////////////////////////////////////////////////////////////////////////
// AssembleFirst: Run first pass of assembler
//                Calculate size info, define labels, do most error checking

void CAssembler::AssembleFirst()
{
	char lbuff[ MAX_LINE_LENGTH ];
	char *tok;			// pointer to token (instruction)
	char *tokdisp = 0;	// pointer to case-sensitive token
	char *arg, *arg2;	// pointer to argument
	int  line = -1;		// current line
	LabelDef lbl, lblnew;
	TableEntry tablinst;

	if (m_pLabelMap != NULL)
		return;
	m_pLabelMap = new CLabelMap;
	m_pLabelMap->InitHashTable(SYMTABLE_HASHSIZE);

	if (m_pLabelDisp != NULL)
		return;
	m_pLabelDisp = new CLabelMap;
	m_pLabelDisp->InitHashTable(SYMTABLE_HASHSIZE);

	m_pOutput->PrintInfo("First pass...\n");
	m_pInterSrc->SeekToBegin();
	m_Loc = 0;

#ifdef _DEBUG
	clock_t clkStart = clock();
#endif

	try {
	 while (m_pInterSrc->ReadString(lbuff, MAX_LINE_LENGTH-1) != NULL) {
		line++;

		if (m_pOutput->CheckCancel())
			m_Status = asmCanceled;
		if (m_Status > asmErrors)
			break;

		// get first token
		tok = strtok( lbuff, szTokSep);
		if (tok == NULL)
			continue;

		if (__iscsymf(lbuff[0]) || lbuff[0] == m_cLocalLabelChar) {	// check for label
			if (tok[strlen(tok) - 1] == ':')
				tok[strlen(tok) - 1] = '\0';

			CheckLabel(tok);

			if(tokdisp) delete[] tokdisp;
			tokdisp = new char[strlen(tok) + 1];
			strcpy(tokdisp, tok);

			if (m_nFlags & chkNoCase)
				strupr(tok);

			// check for .EQU directive
			arg = strtok( NULL, szTokSep);
			if (arg && (!_stricmp(arg, ".EQU") || !_stricmp(arg, "EQU") || arg[0] == '=')) {
				if (arg[0] == '=' && arg[1]) {
					arg++;
					arg2 = strtok( NULL, szLineSep);
					if (arg2)
						arg[strlen(arg)] = ' ';
				}
				else
					arg = strtok( NULL, szLineSep);
				lblnew.islabel = FALSE;
				lblnew.address = EvalExpr( arg, line);
				// check for duplicate label
				if (m_pLabelMap->Lookup( tok, lbl) && lbl.address != lblnew.address)
					AddError( WRN_DUP_LABEL, line, tokdisp);
				// add label
				m_pLabelMap->SetAt( tok, lblnew);
				m_pLabelDisp->SetAt( tokdisp, lblnew);
				continue;
			}
			else {	// add current position label
				// check for instruction with same name
				POSITION pos = m_Table.GetHeadPosition();
				while (pos != NULL) {
					tablinst = m_Table.GetNext( pos);
					if (!_stricmp( tok, tablinst.instr)) {
						AddError( WRN_LABEL_INST, line, tok);
						break;
					}
				}
				lblnew.islabel = TRUE;
				lblnew.address = m_Loc;
				if (m_pLabelMap->Lookup( tok, lbl) && lbl.address != lblnew.address)	// dup label
					AddError( WRN_DUP_LABEL, line, tokdisp);
				m_pLabelMap->SetAt( tok, lblnew);
				m_pLabelDisp->SetAt( tokdisp, lblnew);
			}
			tok = arg;
		}

		// column 1
		if (tok == NULL)
			continue;
		else if (tok[0] == '.') {		// .directive
			// data byte (.db)
			if (!_stricmp(tok, ".DB") || !_stricmp(tok, ".BYTE")) {
				arg = strtok( NULL, szLineSep);
				if (arg == NULL) {
					AddError( ERR_MISSING_EXP, line);
					continue;
				}
				m_pInterSrc->SetLineSecondPass( line);
				char *c = arg;
				while (*c != '\0') {
					while (isspace(*c))
						c++;
					if (*c == '"') {	// string
						c++;
						while (*c != '"' && *c != '\0') {
							if (*c == '\\')	// escape seq
								c++;		// skip backslash
							m_Loc++;
							m_nDataSize++;
							c++;
						}
						if (*c == '\0')
							AddError( ERR_NO_END_QUOTE, line);
						else {
							c++;
							while (isspace(*c))
								c++;
							if (*c == ',')
								c++;
						}
					}
					else if (*c != '\0') {
						while (*c != ',' && *c != '\0')
							c++;
						m_Loc++;
						m_nDataSize++;
						if (*c == ',')
							c++;
					}
				}
			}
			// data word (.dw)
			else if (!_stricmp(tok, ".DW") || !_stricmp(tok, ".WORD")) {
				arg = strtok( NULL, szLineSep);
				if (arg == NULL) {
					AddError( ERR_MISSING_EXP, line);
					continue;
				}
				m_pInterSrc->SetLineSecondPass( line);
				char *c = arg;
				while (*c != '\0') {
					while (isspace(*c))
						c++;
					if (*c != '\0') {
						while (*c != ',' && *c != '\0')
							c++;
						m_Loc += 2;
						m_nDataSize += 2;
						if (*c == ',')
							c++;
					}
				}
			}
			// string data (.text)
			else if (!_stricmp(tok, ".TEXT")) {
				arg = strtok( NULL, szLineSep);
				if (arg == NULL) {
					AddError( ERR_MISSING_STRING, line);
					continue;
				}
				m_pInterSrc->SetLineSecondPass( line);
				char *c = arg;
				while (isspace(*c))
					c++;
				if (*c != '"')
					AddError( ERR_MISSING_STRING, line);
				else {
					c++;
					while (*c != '"' && *c != '\0') {
						if (*c == '\\')	// escape seq
							c++;		// skip backslash
						m_Loc++;
						m_nDataSize++;
						c++;
					}
					if (*c != '"')
						AddError( ERR_NO_END_QUOTE, line);
					m_Loc++;
					m_nDataSize++;
				}
			}
			else if (!_stricmp(tok, ".ORG")) {
				arg = strtok( NULL, szLineSep);
				m_Loc = EvalExpr( arg, line);
				m_pInterSrc->SetLineSecondPass( line);
			}
			else if (!_stricmp(tok, ".PAD")) {
				arg = strtok( NULL, szLineSep);
				int PadAmount = EvalExpr( arg, line) - m_Loc;
				if (PadAmount < 0) {
					AddError( ERR_PAD_NEGATIVE, line);
					continue;
				}
				m_Loc += PadAmount;
				m_nDataSize += PadAmount;
				m_pInterSrc->SetLineSecondPass( line);
			}
			else if (!_stricmp(tok, ".INCBIN")) {
				arg = strtok(NULL, szLineSep);
				if (arg == NULL) {
					AddError( ERR_MISSING_FNAME, line);
					continue;
				}
				CString fn( arg);
				fn.TrimLeft();
				fn.TrimRight();
				// check for quotes
				if (fn.Left(1) == "\"" && fn.Right(1) == "\"")
					fn = fn.Mid( 1, fn.GetLength() - 2);

				CFileStatus fstat;
				// check current directory
				if (!CFile::GetStatus( fn, fstat)) {
					// check include path
					fn = GetOptions()->m_asmIncludePath + "\\" + fn;
					if (!CFile::GetStatus( fn, fstat)) {
						// not found
						AddError( ERR_FILE_NOEXIST, line, arg);
						continue;
					}
				}
				m_pInterSrc->SetLineSecondPass(line);
				m_Loc += (WORD)fstat.m_size;
				m_nDataSize += (WORD)fstat.m_size;
			}
			else if (!_stricmp(tok, ".CHK")) {
				arg = strtok( NULL, szLineSep);
				if (arg == NULL)
					AddError( ERR_MISSING_EXP, line);
				m_pInterSrc->SetLineSecondPass( line);
				m_Loc++;
				m_nDataSize++;
			}
			else if (!_stricmp(tok, ".PLUGIN")) {
				arg = strtok( NULL, szLineSep);
				if (arg == NULL)
					AddError( ERR_PLUGIN_DIR, line);
				else
					m_strPlugin = arg;
			}
			else if (!_stricmp(tok, ".MODULE")) {
				m_pInterSrc->SetLineSecondPass(line);
				arg = strtok( NULL, szTokSep);
				if (arg == NULL)
					AddError( ERR_MODULE_DIR, line);
				else
					m_strModule = arg;
			}
			else if (!_stricmp(tok, ".LOCALLABELCHAR")) {
				m_pInterSrc->SetLineSecondPass(line);
				arg = strtok( NULL, szTokSep);
				if (arg == NULL || arg[0] != '\"' || arg[2] != '\"' || arg[3] != '\0')
					AddError( ERR_LOCALCHAR_DIR, line);
				else
					m_cLocalLabelChar = arg[1];
			}
			else if (!_stricmp(tok, ".ADDINSTR")) {
				arg = strtok( NULL, szLineSep);
				if (!AddTabEntry( arg))
					AddError( ERR_ADDINSTR, line);
			}
			else if (!_stricmp(tok, ".ERROR")) {
				arg = strtok(NULL, szLineSep);
				AddError( ERR_MESSAGE, line, arg);
			}
			else if (!_stricmp(tok, ".WARNING")) {
				arg = strtok(NULL, szLineSep);
				AddError( WRN_MESSAGE, line, arg);
			}
			else if (!_stricmp(tok, ".MESSAGE")) {
				arg = strtok(NULL, szLineSep);
				cout << arg << endl;
			}
			else
				AddError( ERR_UNKNOWN_DIR, line, tok);
		}

		else if (isalpha(tok[0])) {		// instruction
			char *c, *a;
			int valid = -1;	// 0=matching mnemonic found,
							// 1=matching arg format found

			arg = strtok( NULL, szLineSep);
			POSITION nextpos = m_Table.GetHeadPosition();
			POSITION curpos;
			while (nextpos != NULL && valid != 1) {
				curpos = nextpos;
				tablinst = m_Table.GetNext(nextpos);
				if (!_stricmp( tok, tablinst.instr)) {
					valid = 0;
					if (arg == NULL) {
						if (tablinst.arg[0] == '\0') {
							valid = 1;	// matches w/ no args
							m_pInterSrc->SetLineInstruction( line, curpos);
						}
					}
					else {
						c = arg;
						a = tablinst.arg;
						for (;;) {
							while (isspace(*c))
								c++;
							if (*a == '*') {	// variable data
								a++;
								int parendepth = 0;
								bool inquotes = false;
								while (*c != '\0' && (parendepth > 0 || inquotes || *a != toupper(*c)))
								{
									if(*c == '\'')
										inquotes = !inquotes;
									else if(inquotes && *c == '\\')
										c++;
									else if(!inquotes)
										parendepth += (*c == '(') - (*c == ')');
									c++;
								}
								if (*c == '\0' && *a != '\0')
									break;
							}
							else if (*a != toupper(*c))	// doesn't match
								break;
							if (*a == '\0' && *c == '\0') { // entire instruction matches
								valid = 1;
								// remember instruction so we don't have to look
								// it up again on second pass
								m_pInterSrc->SetLineInstruction( line, curpos);
								break;
							}
							c++;
							a++;
						}
					}
				}
			}

			if (valid == -1)
				AddError( ERR_UNKNOWN_INST, line, tok);
			else if (valid == 0)
				AddError( ERR_UNKNOWN_ARG, line, tok);
			else {
				// determine size and advance PC
				m_Loc += tablinst.nbytes;
				m_nCodeSize += tablinst.nbytes;
				m_pInterSrc->SetLineSecondPass( line);
			}
		}
	} }

	catch (CFileException *e) {
		m_Status = asmFileException;
		e->ReportError();
	}
	catch (CMemoryException *e) {
		m_Status = asmMemException;
		e->ReportError();
	}

#ifdef _DEBUG
	float duration = (float)(clock() - clkStart) / CLOCKS_PER_SEC;
	TRACE2("First pass: %f seconds, %d lines per second\n", duration, (int)(m_pInterSrc->GetNumLines() / duration));
#endif

	if (m_nCodeSize + m_nDataSize == 0)
		AddError( ERR_FILE_EMPTY, line);
}


/////////////////////////////////////////////////////////////////////////////
// AssembleMain: Second pass of assembler
//               Evaluate argument data and write object code

void CAssembler::AssembleMain( CFile& binout)
{
	char lbuff[ MAX_LINE_LENGTH ];
	char expr[ MAX_EXPR_LENGTH ];
	int  i;
	char *tok;			// pointer token (instruction)
	char *arg;			// pointer to argument
	WORD argval;		// storage for argument value
	int  line = -1;		// current line
	WORD dwOrg = 0;		// position of start of program (.org directive)
	m_strModule = "";	// name of current module

	ASSERT(m_pLabelMap != NULL);
	ASSERT(m_Status <= asmWarnings);

	m_pOutput->PrintInfo("Second pass...\n");
	m_pInterSrc->SeekToBegin();
	m_Loc = 0;

#ifdef _DEBUG
	clock_t clkStart = clock();
#endif

	try {
	 while (m_pInterSrc->ReadString(lbuff, MAX_LINE_LENGTH-1) != NULL) {
		line++;

		if (m_Status > asmErrors)
			break;
		if (!m_pInterSrc->GetLineSecondPass(line))
			continue;

		// get first token
		tok = strtok( lbuff, szTokSep);
		ASSERT( tok != NULL);

		if (__iscsymf(lbuff[0]) || lbuff[0] == m_cLocalLabelChar) {	// label
			tok = strtok( NULL, szTokSep);
			CheckLabel(tok);

			ASSERT(!tok || _stricmp(tok, ".EQU") || _stricmp(tok, "EQU") || strcmp(tok, "="));
		}

		// column 1
		if (tok == NULL)
			continue;
		else if (tok[0] == '.') {		// .directive
			if (!_stricmp(tok, ".ORG")) {
				arg = strtok( NULL, szLineSep);
				m_Loc = EvalExpr( arg, line);
				dwOrg = m_Loc;
			}
			else if (!_stricmp(tok, ".PAD")) {
				arg = strtok(NULL, szLineSep);
				int PadAmount = EvalExpr( arg, line) - m_Loc;
				char PadWith = 0;
				for (int i = 0; i < PadAmount; i++)
					binout.Write( &PadWith, 1);
				m_Loc += PadAmount;
			}
			else if (!_stricmp(tok, ".MODULE")) {
				m_strModule = strtok( NULL, szLineSep);
			}
			else if (!_stricmp(tok, ".LOCALLABELCHAR")) {
				m_cLocalLabelChar = strtok( NULL, szLineSep)[1];
			}
			// data byte (.db)
			else if (!_stricmp(tok, ".DB") || !_stricmp(tok, ".BYTE")) {
				arg = strtok( NULL, szLineSep);
				ASSERT( arg != NULL);
				char *c = arg;
				while (*c != '\0') {
					while (isspace(*c))
						c++;
					if (*c == '"') {	// string
						c++;
						while (*c != '"' && *c != '\0') {
							if (*c == '\\') {	// escape seq
								c++;			// take next char as literal
								if (*c == '0')
									*c = '\0';	// null char
							}
							binout.Write( c, 1);
							m_Loc++;
							c++;
						}
						ASSERT(*c == '"');
						c++;
						while (isspace(*c))
							c++;
						if (*c == ',')
							c++;
					}
					else if (*c != '\0') {
						for (i = 0; *c != ',' && *c != '\0'; i++) {
							if (i >= MAX_EXPR_LENGTH-1) {
								AddError( ERR_EXPR_TOOLONG, line);
								break;
							}
							expr[i] = *c;
							c++;
						}
						expr[i] = '\0';
						argval = EvalExpr( expr, m_pInterSrc->GetFilePath(line), m_pInterSrc->MapLine(line), TRUE);
						binout.Write( &argval, 1);
						m_Loc++;
						if (*c == ',')
							c++;
					}
				}
			}
			// data word (.dw)
			else if (!_stricmp(tok, ".DW") || !_stricmp(tok, ".WORD")) {
				arg = strtok( NULL, szLineSep);
				ASSERT( arg != NULL);
				char *c = arg;
				while (*c != '\0') {
					while (isspace(*c))
						c++;
					if (*c != '\0') {
						for (i = 0; *c != ',' && *c != '\0'; i++) {
							if (i >= MAX_EXPR_LENGTH-1) {
								AddError( ERR_EXPR_TOOLONG, line);
								break;
							}
							expr[i] = *c;
							c++;
						}
						expr[i] = '\0';
						argval = EvalExpr( expr, line);
						binout.Write( &argval, 2);
						m_Loc += 2;
						if (*c == ',')
							c++;
					}
				}
			}
			// string data (.text)
			else if (!_stricmp(tok, ".TEXT")) {
				arg = strtok( NULL, szLineSep);
				ASSERT( arg != NULL);
				char *c = arg;
				while (isspace(*c))
					c++;
				ASSERT( *c == '"');
				c++;
				while (*c != '"' && *c != '\0') {
					if (*c == '\\') {	// escape seq
						c++;			// take next char as literal
						if (*c == '0')
							*c = '\0';	// null char
					}
					binout.Write( c, 1);
					m_Loc++;
					c++;
				}
				ASSERT(*c == '"');
				*c = '\0';
				binout.Write( c, 1);
				m_Loc++;
			}
			// include binary file (.incbin)
			else if (!_stricmp(tok, ".INCBIN")) {
				arg = strtok(NULL, szLineSep);
				ASSERT(arg != NULL);

				CString fn( arg);
				fn.TrimLeft();
				fn.TrimRight();
				// check for quotes
				if (fn.Left(1) == "\"" && fn.Right(1) == "\"")
					fn = fn.Mid( 1, fn.GetLength() - 2);

				CFileStatus fstat;
				// check current directory
				if (!CFile::GetStatus( fn, fstat)) {
					// check include path
					fn = GetOptions()->m_asmIncludePath + "\\" + fn;
					ASSERT(CFile::GetStatus( fn, fstat));
				}

				BYTE trans;
				CFile fbin( fn, CFile::modeRead);
				for (i = 0; i < (int)fbin.GetLength(); i++) {
					fbin.Read(&trans, 1);
					binout.Write(&trans, 1);
				}
			}
			// checksum byte (.chk)
			else if (!_stricmp(tok, ".CHK")) {
				arg = strtok( NULL, szLineSep);
				LONG fpos = EvalExpr( arg, line) - dwOrg;
				if (fpos < 0 || fpos > (LONG) binout.GetLength())
					AddError( ERR_INVALID_POS, line, arg);
				else {
					binout.Seek( fpos, CFile::begin);
					BYTE bt;
					UINT checksum = 0;
					while (binout.Read(&bt, 1) == 1)
						checksum += bt;
					binout.Write( &checksum, 1);
					m_Loc++;
				}
			}
			else
				ASSERT(0);
		}

		// instruction
		else if (isalpha(tok[0])) {
			char *c, *a;
			int argnum = 0;	// argument number that we're on (1,2...)
			int argbit = 0;	// add this to last byte when done (for ZBIT instructions)
			WORD argsup;	// store value here first
			argval = 0;

			// recall instruction from first pass
			ASSERT( m_pInterSrc->GetLineInstruction(line) != NULL);
			TableEntry tablinst = m_Table.GetAt( m_pInterSrc->GetLineInstruction(line));
			ASSERT( !_stricmp( tok, tablinst.instr));

			arg = strtok( NULL, szLineSep);
			if (arg != NULL) {
				c = arg;
				a = tablinst.arg;
				for (;;) {
					while (isspace(*c))
						c++;
					if (*a == '*') {	// variable data
						argnum++;
						a++;
						int parendepth = 0;
						bool inquotes = false;
						for (i = 0; *c != '\0' && (parendepth > 0 || inquotes || *a != toupper(*c)); i++) {
							if (i >= MAX_EXPR_LENGTH-1) {
								AddError( ERR_EXPR_TOOLONG, line);
								break;
							}
							expr[i] = *c;
							if(*c == '\'')
								inquotes = !inquotes;
							else if(inquotes && *c == '\\')
							{
								c++, i++;
								expr[i] = *c;
							}
							else if(!inquotes)
								parendepth += (*c == '(') - (*c == ')');
							c++;
						}

						if (*c == '\0' && *a != '\0')
							break;

						if(i == 0 && isalpha(*(a-2)))
							expr[i++] = '0';

						expr[i] = '\0';

						// check for apparent illegal indirection
						while (isspace(expr[strlen(expr)-1]))
							expr[strlen(expr)-1] = '\0';
						if ((GetOptions()->m_asmErrorChk & chkIndirect))
						{
							int depth = 0, outsideparens = 0;
							bool parens = false, inquotes = false;
							for(unsigned int k = 0; k < strlen(expr); k++)
							{
								if(expr[k] == '\'')
									inquotes = !inquotes;
								else if(inquotes && expr[k] == '\\')
									k++;
								else if(!inquotes)
								{
									depth += (expr[k] == '(') - (expr[k] == ')');
									if(depth == 0)
										outsideparens++;
									else
										parens = true;
								}
							}

							if(parens && outsideparens == 1)
								AddError( WRN_ILLEGAL_IND, line, expr);
						}

						// evaluate argument
						argsup = EvalExpr(expr, m_pInterSrc->GetFilePath(line), m_pInterSrc->MapLine(line),
							(tablinst.argbytes == 1 && tablinst.modop != TableEntry::OP_R1) ? TRUE : FALSE);

						// perform modops
						if (tablinst.modop == TableEntry::OP_R1) {
							// relative address (1 byte)
							argval = (BYTE)(argsup - m_Loc - 2);
							if (argsup - m_Loc - 2 > 127 || argsup - m_Loc - 2 < -128)
								AddError( ERR_REL_JUMP, line);
						}
						else if (argnum == 1 && tablinst.modop == TableEntry::OP_ZBIT) {
							// bit num argument
							argbit = argsup;
						}
						else if (argnum == 2 && tablinst.modop == TableEntry::OP_COMBINE) {
							// two args
							if ((short)argval > 255 || (short)argval < -128 ||
								(short)argsup > 255 || (short)argsup < -128)
								AddError( WRN_MSBYTE_UNUSED, line);
							argval &= 0x00FF;
							argval += (argsup << 8);
						}
						else if (argnum == 2 && tablinst.modop == TableEntry::OP_CSWAP) {
							// two args, swapped
							if ((short)argval > 127 || (short)argval < -128 ||
								(short)argsup > 127 || (short)argsup < -128)
								AddError( WRN_MSBYTE_UNUSED, line);
							argval = argval << 8;
							argval += (argsup & 0x00FF);
						}
						else
							argval = argsup;
					}

					// everything should match
					ASSERT (*a == toupper(*c));
					if (*a == '\0' && *c == '\0')	// end of instruction
						break;
					c++;
					a++;
				}
			}

			// carry out post-manipulations
			argval |= tablinst.or;
			if (argbit) {
				if (argbit > 7)
					AddError( ERR_INVALID_BIT, line, (char *)argbit);
				argbit *= 8;
				if (tablinst.nbytes <= 2)
					tablinst.opcode += (argbit << 8);
				else if (tablinst.nbytes == 3)
					argval += argbit;
				else if (tablinst.nbytes == 4)
					argval += (argbit << 8);
			}

			// write the object code
			binout.Write( &tablinst.opcode, tablinst.opbytes);
			if (tablinst.argbytes > 0)
				binout.Write( &argval, tablinst.argbytes);
			m_Loc += tablinst.nbytes;	// advance PC
		}
	} }

	catch (CFileException *e) {
		m_Status = asmFileException;
		e->ReportError();
	}
	catch (CMemoryException *e) {
		m_Status = asmMemException;
		e->ReportError();
	}

#ifdef _DEBUG
	float duration = (float)(clock() - clkStart) / CLOCKS_PER_SEC;
	TRACE2("Second pass: %f seconds, %d lines per second\n", duration, (int)(m_pInterSrc->GetNumLines() / duration));
#endif
}

/////////////////////////////////////////////////////////////////////////////
// EvalExpr
//   Evaluate numeric expression <expr>
//   can consist of constants, labels, operators
//    <filename> and <line> are for error output
//    <bByte> = TRUE if result should be single byte

static char szOperandSep[] = " \t+-*/=<>!&|^()";

// absolute line number version
inline WORD CAssembler::EvalExpr(char *expr, int absLine)
	{ return EvalExpr( expr, m_pInterSrc->GetFilePath(absLine), m_pInterSrc->MapLine(absLine)); }


enum OPERATOR {
	opNone = 0, opPlus, opMinus, opNot,	/* <-- can be unary */
	opMult, opDivide,
	opEqual, opNEqual, opLess, opGreater, opLessE, opGreaterE,
	opAnd, opOr, opXor, opShiftL, opShiftR
};

WORD CAssembler::EvalExpr( char *expr, CString filename, int line, BOOL bByte)
{
	ASSERT(!strchr(expr, ';'));
	if (expr == NULL)
	{
		AddError( ERR_MISSING_EXP, filename, line);
		return 0;
	}
	m_cOriginalExpr = expr;
	char *exprbuffer = new char[strlen(expr) + 1];
	strcpy(exprbuffer, expr);
	m_cExpr = exprbuffer;
	m_sFilename = filename;
	m_nLine = line;
	m_bValidExpr = true;

	while(isspace(m_cExpr[strlen(m_cExpr) - 1]))
		m_cExpr[strlen(m_cExpr) - 1] = '\0';

	int result = EvalBoolean();

	SkipSpace();

	if(m_bValidExpr && *m_cExpr != '\0')
		AddError( ERR_ILLEGAL_EXPR, filename, line, m_cOriginalExpr);

	delete[] exprbuffer;

	if(bByte && result > 255 && result < 65535)
		AddError( WRN_MSBYTE_UNUSED, filename, line, m_cOriginalExpr);
	else if(result > 65535)
		AddError( ERR_OVERFLOW, filename, line, m_cOriginalExpr);

	return (WORD)result;
}

void CAssembler::SkipSpace()
{
	while(isspace(*m_cExpr))
		m_cExpr++;
}

int CAssembler::EvalBoolean()
{
	int result = 0;
	OPERATOR nOperator = opPlus;

	while(m_bValidExpr && nOperator != opNone)
	{
		int result2 = EvalBitLogic();

		switch(nOperator)
		{
		case opEqual:
			result = (result == result2);
			break;

		case opNEqual:
			result = (result != result2);
			break;

		case opLess:
			result = (result < result2);
			break;

		case opGreater:
			result = (result < result2);
			break;

		case opLessE:
			result = (result <= result2);
			break;

		case opGreaterE:
			result = (result >= result2);
			break;

		case opPlus:
			result += result2;
			break;
		}

		nOperator = opNone;
		SkipSpace();

		switch(*m_cExpr++)
		{
		case '>':
			if(*m_cExpr++ == '=')
				nOperator = opGreaterE;
			else
			{
				nOperator = opGreater;
				m_cExpr--;
			}
			break;

		case '<':
			if(*m_cExpr++ == '=')
				nOperator = opLessE;
			else
			{
				nOperator = opLess;
				m_cExpr--;
			}
			break;

		case '!':
			if(*m_cExpr++ == '=')
				nOperator = opNEqual;
			else
				m_cExpr -= 2;
			break;

		case '=':
			if(*m_cExpr++ == '=')
				nOperator = opEqual;
			else
				m_cExpr -= 2;
			break;

		default:
			m_cExpr--;
		}
	}

	return result;
}

int CAssembler::EvalBitLogic()
{
	int result = 0;
	OPERATOR nOperator = opOr;

	while(m_bValidExpr && nOperator != opNone)
	{
		int result2 = EvalAddition();

		switch(nOperator)
		{
		case opAnd:
			result &= result2;
			break;

		case opOr:
			result |= result2;
			break;

		case opXor:
			result ^= result2;
			break;

		case opShiftL:
			result = (result << result2);
			break;

		case opShiftR:
			result = (result >> result2);
			break;
		}

		nOperator = opNone;
		SkipSpace();

		switch(*m_cExpr++)
		{
		case '&':
			nOperator = opAnd;
			break;

		case '|':
			nOperator = opOr;
			break;

		case '^':
			nOperator = opXor;
			break;

		case '<':
			if(*m_cExpr++ == '<')
				nOperator = opShiftL;
			else
				m_cExpr -= 2;
			break;

		case '>':
			if(*m_cExpr++ == '>')
				nOperator = opShiftR;
			else
				m_cExpr -= 2;
			break;

		default:
			m_cExpr--;
		}
	}

	return result;
}


int CAssembler::EvalAddition()
{
	int result = 0;
	OPERATOR nOperator = opPlus;

	while(m_bValidExpr && nOperator != opNone)
	{
		int result2 = EvalProduct();

		switch(nOperator)
		{
		case opPlus:
			result += result2;
			break;

		case opMinus:
			result -= result2;
			break;
		}

		nOperator = opNone;
		SkipSpace();

		switch(*m_cExpr++)
		{
		case '+':
			nOperator = opPlus;
			break;

		case '-':
			nOperator = opMinus;
			break;

		default:
			m_cExpr--;
		}
	}

	return result;
}


int CAssembler::EvalProduct()
{
	int result = 1;
	OPERATOR nOperator = opMult;

	while(m_bValidExpr && nOperator != opNone)
	{
		int result2 = EvalNumber();

		switch(nOperator)
		{
		case opMult:
			result *= result2;
			break;

		case opDivide:
			if (result2 == 0)
			{
				AddError( ERR_DIVIDEBYZERO, m_sFilename, m_nLine);
				m_bValidExpr = false;
			}
			else
				result /= result2;
			break;
		}

		nOperator = opNone;
		SkipSpace();

		switch(*m_cExpr++)
		{
		case '(':
			m_cExpr--;

		case '*':
			nOperator = opMult;
			break;

		case '/':
			nOperator = opDivide;
			break;

		default:
			m_cExpr--;
		}
	}

	return result;
}

int CAssembler::EvalNumber()
{
	char labeldisp[MAX_EXPR_LENGTH];
	int result = 0;

	OPERATOR nOperator = opNone;
	SkipSpace();

	switch(*m_cExpr++)
	{
	case '+':
		break;

	case '-':
		nOperator = opMinus;
		break;

	case '!':
		nOperator = opNot;

	default:
		m_cExpr--;
	}

	SkipSpace();

	if(*m_cExpr == '(')
	{
		m_cExpr++;
		// subexpression
		result = EvalBoolean();

		SkipSpace();

		if(m_bValidExpr && *m_cExpr != ')')
		{
			AddError( ERR_NO_CLOSEPARENTH, m_sFilename, m_nLine);
			m_bValidExpr = false;
			return result;
		}

		m_cExpr++;
	}
	else
	{
		// get numeric expression

		if (*m_cExpr == '\0')
		{
			AddError(ERR_MISSING_VALUE, m_sFilename, m_nLine);
			m_bValidExpr = false;
		}
		// numeric constant

		// hexadecimal
		else if (*m_cExpr == '$' && isxdigit(m_cExpr[1])
				|| GetLastHexChar() == 'h' && isdigit(*m_cExpr))
		{
			if(*m_cExpr == '$') m_cExpr++;

			while(isxdigit(*m_cExpr))
			{
				result = result * 16 + 
					((*m_cExpr >= '0' && *m_cExpr <= '9')
						? (*m_cExpr - '0')
						: (toupper(*m_cExpr) - 'A' + 10));
				m_cExpr++;
			}

			if(tolower(*m_cExpr) == 'h') m_cExpr++;
		}

		// binary
		else if (*m_cExpr == '%' && (m_cExpr[1] == '0' || m_cExpr[1] == '1')
				|| GetLastBinChar() == 'b' && (*m_cExpr == '0' || *m_cExpr == '1'))
		{
			if(*m_cExpr == '%') m_cExpr++;

			while(*m_cExpr == '0' || *m_cExpr == '1')
				result = (result << 1) + (*m_cExpr++ - '0');

			if(tolower(*m_cExpr) == 'b') m_cExpr++;
		}

		// decimal
		else if (isdigit(*m_cExpr))
		{
			do
				result = 10 * result + *m_cExpr - '0';
			while(isdigit(*++m_cExpr));
		}

		// character constant
		else if (*m_cExpr == '\'')
		{
			m_cExpr++;
			if(*m_cExpr == '\\') m_cExpr++;
			result = *m_cExpr++;
			if(*m_cExpr == '\'')
				m_cExpr++;
			else
			{
				AddError( ERR_NO_END_QUOTE, m_sFilename, m_nLine);
				m_bValidExpr = false;
			}
		}

		// label
		else if (__iscsymf(*m_cExpr) || *m_cExpr == m_cLocalLabelChar)
		{
			char *label = new char[MAX_EXPR_LENGTH];

			for(int i = 0; (__iscsym(*m_cExpr) || *m_cExpr == m_cLocalLabelChar); i++)
				label[i] = *m_cExpr++;
			
			label[i] = '\0';

			CheckLabel(label);

			strcpy(labeldisp, label);

			if (m_nFlags & chkNoCase)
				strupr(label);
			LabelDef lbl;
			if (!m_pLabelMap->Lookup( label, lbl))
			{
				AddError( ERR_UNKNOWN_LABEL, m_sFilename, m_nLine, labeldisp);
				m_bValidExpr = false;
				lbl.islabel = 0;
				lbl.address = 0;
				m_pLabelMap->SetAt(label, lbl);
			}
			result = lbl.address;
		}

		// location counter
		else if (*m_cExpr == '$')
		{
			m_cExpr++;
			result = m_Loc;
		}

		// unknown
		else
		{
			AddError( ERR_ILLEGAL_EXPR, m_sFilename, m_nLine, m_cOriginalExpr);
			m_bValidExpr = false;
		}
	}

	switch(nOperator)
	{
	case opMinus:
		result = -result;
		break;

	case opNot:
		result = !result;
		break;
	}

	return result;
}

char CAssembler::GetLastHexChar()
{
	for(int i = 0; isxdigit(m_cExpr[i]); i++);
	return tolower(m_cExpr[i]);
}

char CAssembler::GetLastBinChar()
{
	for(int i = 0; m_cExpr[i] == '0' || m_cExpr[i] == '1'; i++);
	return tolower(m_cExpr[i]);
}


/////////////////////////////////////////////////////////////////////////////
// AddError
//   Add an error to the error list with specified attributes

void CAssembler::AddError(int errcode, int absLine, const char *sArg)
{
	AddError( errcode, m_pInterSrc->GetFilePath(absLine), m_pInterSrc->MapLine(absLine), sArg);
}

void CAssembler::AddError(int errcode, CString filename, int line, const char *sArg)
{
	CAssemblyError err;
	err.m_ErrCode = errcode;
	err.m_FileName = filename;
	err.m_LineNum = line;
	err.m_bWarning = FALSE;

	switch (errcode) {

	// Miscellaneous/system errors
#ifdef _DEBUG
	case ERR_TEST:
		err.m_Message = "Testing error (debug)";
		break;
	case ERR_UNKNOWN:
		err.m_Message = "Unexpected results while assembling";
		break;
#endif
	case ERR_MESSAGE:
		err.m_Message = sArg;
		break;
	case WRN_MESSAGE:
		err.m_Message = sArg;
		err.m_bWarning = TRUE;
		break;
	case ERR_FILE_EMPTY:
		err.m_Message = "Output file is empty!";
		break;
	case ERR_PLUGIN:
		err.m_Message.Format("Plugin file \"%s\" not found", sArg);
		break;

	// Preprocessor errors
	case ERR_FILE_OPEN:
		err.m_Message.Format("Error opening file %s", sArg);
		break;
	case ERR_FILE_NOEXIST:
		err.m_Message.Format("Cannot find file %s", sArg);
		break;
	case WRN_FILE_INCLUDED:
		err.m_Message.Format("File %s included more than once", sArg);
		err.m_bWarning = TRUE;
		break;
	case ERR_LINE_TOOLONG:
		err.m_Message = "Line is too long (>255 chars)";
		break;
	case ERR_UNKNOWN_PRE_DIR:
		err.m_Message.Format("Unrecognized preprocessor directive: \"%s\"", sArg);
		break;
	case ERR_MISSING_FNAME:
		err.m_Message = "Filename expected";
		break;
	case ERR_MISSING_MACRO:
		err.m_Message = "Macro name expected";
		break;
	case ERR_TOOMANY_ARGS:
		err.m_Message.Format("No more than 10 arguments allowed for macro %s", sArg);
		break;
	case ERR_INVALID_ARGS:
		err.m_Message.Format("Invalid argument format for macro %s", sArg);
		break;
	case ERR_ARG_COUNT:
		err.m_Message.Format("Wrong number of arguments for macro %s", sArg);
		break;
	case ERR_NO_ENDIF:
		err.m_Message = "End of file before #endif directive";
		break;
	case ERR_ELSE_NOIF:
		err.m_Message = "#else directive without #if";
		break;
	case ERR_ENDIF_NOIF:
		err.m_Message = "#endif directive without #if";
		break;

	// Parsing errors
	case ERR_MISSING_EXP:
		err.m_Message = "Value or expression expected";
		break;
	case ERR_UNKNOWN_LABEL:
		err.m_Message.Format("Label not defined: \"%s\"", sArg);
		break;
	case ERR_ILLEGAL_EXPR:
		err.m_Message.Format("Illegal expression: \"%s\"", sArg);
		break;
	case ERR_NO_END_QUOTE:
		err.m_Message = "End of line in quoted string";
		break;
	case ERR_MISSING_VALUE:
		err.m_Message = "Value expected";
		break;
	case ERR_MISSING_STRING:
		err.m_Message = "Quoted string expected";
		break;
	case ERR_DIVIDEBYZERO:
		err.m_Message = "Divide by zero";
		break;
	case ERR_NO_OPENPARENTH:
		err.m_Message = "Missing opening '('";
		break;
	case ERR_NO_CLOSEPARENTH:
		err.m_Message = "Missing closing ')'";
		break;
	case ERR_EXPR_TOOLONG:
		err.m_Message = "Expression is too long";
		break;

	// Label/directive errors
	case ERR_UNKNOWN_DIR:
		err.m_Message.Format("Unrecognized directive: \"%s\"", sArg);
		break;
	case WRN_DUP_LABEL:
		err.m_Message.Format("Duplicate label \"%s\"", sArg);
		err.m_bWarning = TRUE;
		break;
	case WRN_LABEL_INST:
		err.m_Message.Format("Whitespace needed before instruction \"%s\"", sArg);
		err.m_bWarning = TRUE;
		break;
	case ERR_INVALID_POS:
		err.m_Message.Format("Illegal position value: %s", sArg);
		break;
	case ERR_ADDINSTR:
		err.m_Message = "Invalid .ADDINSTR directive";
		break;
	case ERR_PAD_NEGATIVE:
		err.m_Message = "Cannot pad backwards";
		break;
	case ERR_PLUGIN_DIR:
		err.m_Message = "Plugin name expected";
		break;
	case ERR_MODULE_DIR:
		err.m_Message = "Module name expected";
		break;
	case ERR_LOCALCHAR_DIR:
		err.m_Message = "Local label char must be surrounded by \"\"";
		break;

	// Instruction table errors
	case ERR_UNKNOWN_INST:
		err.m_Message.Format("Unrecognized instruction: \"%s\"", sArg);
		break;
	case ERR_UNKNOWN_ARG:
		err.m_Message.Format("Unrecognized argument format for %s instruction", sArg);
		break;
	case ERR_MISSING_DATA:
		err.m_Message = "Missing data in argument";
		break;
	case ERR_REL_JUMP:
		err.m_Message = "Relative jump too long";
		break;
	case ERR_INVALID_BIT:
		err.m_Message.Format("Illegal bit value '%d'", (WORD) sArg);
		break;
	case ERR_OVERFLOW:
		err.m_Message = "Value overflows word";
		break;
	case WRN_MSBYTE_UNUSED:
		err.m_Message = "Unused data in MS byte of argument";
		err.m_bWarning = TRUE;
		break;
	case WRN_ILLEGAL_IND:
		err.m_Message.Format("Apparent illegal indirection: \"%s\"", sArg);
		err.m_bWarning = TRUE;
		break;

	default:
		err.m_Message = "[Unknown error code]";
		TRACE1("Unknown error code %d!\n", errcode);
		break;
	}

	if (err.m_bWarning) {
		m_nWarnings++;
		if (m_Status < asmWarnings)
			m_Status = asmWarnings;
	}
	else {
		m_nErrors++;
		if (m_Status < asmErrors)
			m_Status = asmErrors;
	}

	m_pOutput->PrintError(&err);
}


/////////////////////////////////////////////////////////////////////////////
// Finish: show finishing stuff

void CAssembler::Finish() const
{
	if (m_Status <= asmErrors) {
		CString msg;
		msg.Format("Assembly complete - %d error(s), %d warning(s)\n", m_nErrors, m_nWarnings);
		m_pOutput->PrintInfo(msg);
		m_pOutput->PrintInfo("");
	}
	else if (m_Status == asmCanceled)
		m_pOutput->PrintInfo("Assembly canceled by user\n");
	else if (m_Status == asmFileException)
		m_pOutput->PrintInfo("File I/O error, aborting...\n");
	else if (m_Status == asmMemException)
		m_pOutput->PrintInfo("Out of memory, aborting...\n");

//	if (m_Status == asmOK)
//		MessageBeep( MB_ICONASTERISK);
	if (m_Status == asmWarnings)
		MessageBeep( MB_ICONEXCLAMATION);
	else if (m_Status != asmOK)
		MessageBeep( MB_ICONSTOP);
}

/////////////////////////////////////////////////////////////////////////////
// CheckLabel: handle local labels

void CAssembler::CheckLabel(char* &tok)
{
	if(tok[0] != m_cLocalLabelChar)
		return;

	char *newtok = new char[strlen((LPCSTR)m_strModule) + strlen(tok) + 1];
	strcpy(newtok, (LPCSTR)m_strModule);
	strcat(newtok, tok);
	tok = newtok;
}