/////////////////////////////////////////////////////////////////////////////
// Assembler.h : interface for assembler base class
//
// Copyright (C) 2000 Jeremy Goetsch (jeremy@acz.org)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//

#if !defined(AFX_ASSEMBLER_H__6F6244C2_546B_11D1_8E22_90370EDCC92D__INCLUDED_)
#define AFX_ASSEMBLER_H__6F6244C2_546B_11D1_8E22_90370EDCC92D__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

/////////////////////////////////////////////////////////////////////////////
// Assembler Class Hierarchy:
//    CAssembler - abstract
//       CTwoPassAssembler - abstract
//          CPluginAssembler
//          CHexAssembler
//       COnePassAssembler - abstract
//          CStatsAssembler
//          CSymTableAssembler

// maximums
#define MAX_MACRO_ARGS		10
#define MAX_NEST_LEVELS		31

// Table data format
//
struct TableEntry
{
	char  instr[8];		// instruction mnemonic
	char  arg[16];		// argument format
	DWORD opcode;		// opcode bytes
	short opbytes;		// length of opcode
	short argbytes;		// length of args
	short nbytes;		// total length (opcode + args)
	enum {
		OP_NOTOUCH,		// no changes
		OP_R1,			// 8-bit relative address
		OP_ZBIT,		// Z80 bit instruction
		OP_COMBINE,		// combine LS bytes of first two args
		OP_CSWAP,		// combine and swap
	} modop;
	BYTE tclass;		// instruction set class (not implemented)
	WORD shift;			// shift first arg left (not implemented)
	WORD or;			// or mask for first arg
};

/////////////////////////////////////////////////////////////////////////////
// CAssemblyError class

class CAssemblyError
{
public:
	BOOL m_bWarning;
	CString m_FileName;
	int m_LineNum;
	UINT m_ErrCode;
	CString m_Message;
};

class CMacroDef
{
public:
	CMacroDef();
	CMacroDef( LPCSTR szDefinition);
	void Empty();

	int m_NumArgs;
	CString m_szArgs[MAX_MACRO_ARGS];
	CString m_szDefinition;
};

typedef CMap<CString,LPCSTR,CMacroDef,CMacroDef&> CMacroMap;

struct LabelDef
{
	char islabel;	// nonzero if position label, zero if equate
	WORD address;
};

typedef CMap<CString,LPCSTR,LabelDef,LabelDef&> CLabelMap;

// forward declares
class CAsmOutputStream;
class CIntermediateFile;

// assembler option flags
enum {
	chkNone			= 0x0000,
	chkIndirect		= 0x0001,
	chkNoCase		= 0x0002,
	chkDefault		= chkIndirect
};

/////////////////////////////////////////////////////////////////////////////
// CAssembler abstract base class

class CAssembler
{
public:
	CAssembler(UINT nFlags, LPCSTR szDefines = NULL);
	virtual ~CAssembler();

// Operations
public:
	BOOL LoadTabFile( CString filename);
	//void SetOutput(CAsmOutputStream *pOutput);

	// Overridable build function
	virtual void Build( CString filename) = 0;

	// Transfers the first pass info from a previous assembly
	virtual void LoadFirstPass( CAssembler *pSourceAsm);

	// Output finishing text and sound
	virtual void Finish() const;

	enum ASMSTATUS {	// status
		asmOK,				// everything kosher
		asmWarnings,		// warnings are present
		asmErrors,			// errors are present
		asmCanceled,		// assembly canceled
		asmFileException,	// file i/o error
		asmMemException,	// out of memory
	};
	ASMSTATUS GetStatus() const { return m_Status; }

// Implementaion
protected:
	BOOL AddTabEntry( char *EntryLine);
	virtual void Preprocess( CString filename, CMacroMap* pMacroMap = NULL);
	virtual void AssembleFirst();
	virtual void AssembleMain( CFile& binout);
	void CheckLabel( char* &tok );
	void AddError( int errcode, int absLine, const char *sArg = NULL);
	void AddError( int errcode, CString filename, int line, const char *sArg = NULL);
	WORD EvalExpr( char *expr, CString filename, int line, BOOL bByte = FALSE);

	void SkipSpace();
	int EvalBitLogic();
	int EvalBoolean();
	int EvalAddition();
	int EvalProduct();
	int EvalNumber();

	inline char GetLastHexChar();
	inline char GetLastBinChar();

	CAsmOutputStream *m_pOutput;

	UINT m_nFlags;
	LPCSTR m_pszDefines;
	CString m_strPlugin;
	CString m_strModule;
	char m_cLocalLabelChar;

	char *m_cExpr,*m_cOriginalExpr;
	CString m_sFilename;
	int m_nLine;
	boolean m_bValidExpr;

	// instuction set definition table
	CList<TableEntry,TableEntry&> m_Table;
	// label definition table
	CLabelMap *m_pLabelMap;
	// labels for display (case sensitive)
	CLabelMap *m_pLabelDisp;
	CIntermediateFile *m_pInterSrc;
	WORD m_Loc;			// the byte location counter (PC)

	// assembly status
	ASMSTATUS m_Status;
	int  m_nErrors;		// number of errors
	int  m_nWarnings;	// number of warnings

	int  m_numFiles;	// number of files assembled
	int  m_nCodeSize;	// number of bytes of code
	int  m_nDataSize;	// number of bytes of data

private:
	WORD EvalExpr( char *expr, int absLine);
};

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_ASSEMBLER_H__6F6244C2_546B_11D1_8E22_90370EDCC92D__INCLUDED_)
